﻿using System;
using System.Collections.Generic;

namespace MiniClousot
{
  public abstract class AST
  {
  }

  #region Expressions
  public enum BinaryOp { Add, Sub, Mul, Div, And, Or, LessEqual, Less, Equal, GreaterEqual, Greater, NotEqual };
  public enum UnaryOp { Minus, Not };

  public abstract class Expression : AST
  {
    virtual public bool IsBinary
    {
      get
      {
        return false;
      }
    }

    virtual public bool IsUnary
    {
      get
      {
        return false;
      }
    }

    virtual public bool IsVar
    {
      get
      {
        return false;
      }
    }

    virtual public bool IsSlackVar
    {
      get
      {
        return false;
      }
    }

    virtual public bool IsConst
    {
      get
      {
        return false;
      }
    }

    virtual public Set<Expression> Variables
    {
      get
      {
        return new Set<Expression>();
      }
    }

    virtual public void Unpack(out BinaryOp op, out Expression left, out Expression right)
    {
      throw new InvalidOperationException();
    }

    virtual public void Unpack(out UnaryOp op, out Expression left)
    {
      throw new InvalidOperationException();
    }

    virtual public void Unpack(out int value)
    {
      throw new InvalidOperationException();
    }
  }

  public class Const : Expression
  {
    int k;
    public Const(int k)
    {
      this.k = k;
    }

    public override bool IsConst
    {
      get
      {
        return true;
      }
    }

    public override void Unpack(out int value)
    {
      value = this.k;
    }

    public override int GetHashCode()
    {
      return k;
    }

    public override bool Equals(object obj)
    {
      Const objAsConst = obj as Const;

      if (objAsConst == null)
        return false;

      return this.k == objAsConst.k;
    }

    public override string ToString()
    {
      return k.ToString();
    }
  }

  public class Var : Expression
  {
    string name;

    public Var(string name)
    {
      this.name = name;
    }

    public override bool IsVar
    {
      get
      {
        return true;
      }
    }

    public override Set<Expression> Variables
    {
      get
      {
        return new Set<Expression>(this);
      }
    } 

    public override int GetHashCode()
    {
      return name.GetHashCode();
    }

    public override bool Equals(object obj)
    {
      Var objAsVar = obj as Var;
      if (objAsVar == null)
        return false;

      return this.name == objAsVar.name;
    }

    public override string ToString()
    {
      return this.name;
    }
  }

  public class SlackVar : Var
  {
    static int counter = 0;

    public SlackVar()
      : base ("skv" + (counter++).ToString())
    {
    }

    public override bool IsSlackVar
    {
      get
      {
        return true;
      }
    }
  }

  public class UnaryExpression : Expression
  {
    UnaryOp op;
    Expression exp;

    public UnaryExpression(UnaryOp op, Expression exp)
    {
      this.op = op;
      this.exp = exp;
    }

    public override bool IsUnary
    {
      get
      {
        return true;
      }
    }

    public override void Unpack(out UnaryOp op, out Expression exp)
    {
      op = this.op;
      exp = this.exp;
    }

    public override Set<Expression> Variables
    {
      get
      {
        return this.exp.Variables ;
      }
    }

    public override string ToString()
    {
      return string.Format("{0}({1})", this.op, this.exp);
    }
  }

  public class BinaryExpression : Expression
  {
    BinaryOp op;
    Expression left, right;

    public BinaryExpression(BinaryOp op, Expression left, Expression right)
    {
      this.op = op;
      this.left = left;   
      this.right = right;
    }

    public override bool IsBinary
    {
      get
      {
        return true;
      }
    }

    public override void Unpack(out BinaryOp op, out Expression left, out Expression right)
    {
      op = this.op;
      left = this.left;
      right = this.right;
    }

    public override Set<Expression> Variables
    {
      get
      {
        return (Set<Expression>) this.left.Variables.Union(this.right.Variables);
      }
    }

    public override string ToString()
    {
      return string.Format("{0}({1}, {2})", this.op, this.left, this.right);
    }
  }

  #endregion

  #region Statements
  public abstract class Statement : AST
  {
    virtual public void Unpack(out Var x, out Expression exp)
    {
      throw new InvalidOperationException();
    }

    virtual public void Unpack(out Statement s1, out Statement s2)
    {
      throw new InvalidOperationException();
    }

    virtual public void Unpack(out Expression exp, out Statement s1, out Statement s2)
    {
      throw new InvalidOperationException();
    }

    virtual public void Unpack(out Expression exp, out Statement body)
    {
      throw new InvalidOperationException();
    }

    virtual public void Unpack(out Expression exp)
    {
      throw new InvalidOperationException();
    }
  }

  public class Nop : Statement 
  {
    public Nop()
    {
    }

    public override bool Equals(object obj)
    {
      return obj is Nop;
    }

    public override string ToString()
    {
      return "nop";
    }

    public override int GetHashCode()
    {
      return "nop".GetHashCode();
    }
  }

  public class Assign : Statement
  {
    Var x;
    Expression exp;

    public Assign(Var x, Expression exp)
    {
      this.x = x;
      this.exp = exp;
    }

    override public void Unpack(out Var x, out Expression exp)
    {
      x = this.x;
      exp = this.exp;
    }

    public override string ToString()
    {
      return string.Format("{0} := {1}", this.x, this.exp);
    }
  }

  public class Seq : Statement
  {
    Statement s1, s2;

    public Seq(Statement s1, Statement s2)
    {
      this.s1 = s1;
      this.s2 = s2;
    }

    public override void Unpack(out Statement s1, out Statement s2)
    {
      s1 = this.s1;
      s2 = this.s2;
    }

    public override string ToString()
    {
      return string.Format(";({0}, {1})", this.s1, this.s2);
    }
  }

  public class If : Statement
  {
    Expression cond;
    Statement s1, s2;

    public If(Expression cond, Statement s1, Statement s2)
    {
      this.cond = cond;
      this.s1 = s1;
      this.s2 = s2;
    }

    public override void Unpack(out Expression exp, out Statement s1, out Statement s2)
    {
      exp = this.cond;
      s1 = this.s1;
      s2 = this.s2;
    }

    public override string ToString()
    {
      return string.Format("If({0}, {1}, {2})", this.cond, this.s1, this.s2);
    }
  }

  public class While : Statement
  {
    Expression cond;
    Statement body;

    public While(Expression cond, Statement body)
    {
      this.cond = cond;
      this.body = body;
    }

    public override void Unpack(out Expression exp, out Statement body)
    {
      exp = this.cond;
      body = this.body;
    }

    public override string ToString()
    {
      return string.Format("While({0}, {1})", this.cond, this.body);
    }
  }

  public class Assert : Statement
  {
    Expression cond;

    public Assert(Expression cond)
    {
      this.cond = cond;
    }

    public override void Unpack(out Expression exp)
    {
      exp = this.cond;
    }

    public override string ToString()
    {
      return string.Format("Assert({0})", this.cond.ToString());
    }
  }

  public class Assume : Statement
  {
    Expression cond;

    public Assume(Expression cond)
    {
      this.cond = cond;
    }

    public override void Unpack(out Expression exp)
    {
      exp = this.cond;
    }

    public override string ToString()
    {
      return string.Format("Assume({0})", this.cond.ToString());
    }
  }
#endregion

  #region Statement Visitor
  abstract public class Visitor<Out>
  {
    protected Out result;

    public Out Result
    {
      get
      {
        return this.result;
      }
    }

    public void Visit(Statement s)
    {
      Var x;
      Expression exp; 
      Statement s1, s2;
      if (s is Nop)
      {
        VisitNop();
      }
      else if (s is Assign)
      {
        s.Unpack(out x, out exp);
        VisitAssign(x, exp);
      }
      else if (s is Seq)
      {
        s.Unpack(out s1, out s2);
        VisitSeq(s1, s2);
      }
      else if (s is If)
      {
        s.Unpack(out exp, out s1, out s2);
          VisitIf(exp, s1, s2);
      }
      else if (s is While)
      {
        s.Unpack(out exp, out s1);
        VisitWhile(exp, s1);
      }
      else if (s is Assert)
      {
        s.Unpack(out exp);
        VisitAssert(exp);
      }
      else if (s is Assume)
      {
        s.Unpack(out exp);
        VisitAssume(exp);
      }
    }

    virtual protected void VisitSeq(Statement s1, Statement s2)
    {
    }

    virtual protected void VisitAssume(Expression exp)
    { 
    }

    virtual protected void VisitAssert(Expression exp)
    {
    }

    virtual protected void VisitIf(Expression exp, Statement s1, Statement s2)
    {
    }

    virtual protected void VisitWhile(Expression exp, Statement s1)
    {
    }

    virtual protected void VisitAssign(Var x, Expression exp)
    {      
    }

    virtual protected void VisitNop()
    {
    }

  }
  #endregion
}
