﻿using System;
using Microsoft.Research.AbstractDomains.Expressions;

namespace MiniClousot
{
  class Decoder : IExpressionDecoder<Expression>
  {
    #region IExpressionDecoder<Expression> Members

    public ExpressionOperator OperatorFor(Expression exp)
    {
      UnaryOp un_op; BinaryOp bin_op;
      Expression exp1, exp2;
      if (exp.IsBinary)
      {
        exp.Unpack(out bin_op, out exp1, out exp2);
        return ToExpressionOperator(bin_op);
      }
      if (exp.IsUnary)
      {
        exp.Unpack(out un_op, out exp1);
        return ToExpressionOperator(un_op);
      }
      if (exp.IsConst)
      {
        return ExpressionOperator.Constant;
      }
      if (exp.IsVar)
      {
        return ExpressionOperator.Variable;
      }

      throw new Exception();
    }

    public Expression LeftExpressionFor(Expression exp)
    {
      UnaryOp un_op; BinaryOp bin_op;
      Expression exp1, exp2;
      if (exp.IsBinary)
      {
        exp.Unpack(out bin_op, out exp1, out exp2);
        return exp1;
      }
      if (exp.IsUnary)
      {
        exp.Unpack(out un_op, out exp1);
        return exp1;
      }

      throw new Exception();
    }

    public Expression RightExpressionFor(Expression exp)
    {
      BinaryOp bin_op;
      Expression exp1, exp2;
      if (exp.IsBinary)
      {
        exp.Unpack(out bin_op, out exp1, out exp2);
        return exp2;
      }
      throw new Exception();
    }

    public Expression Stripped(Expression exp)
    {
      return exp;
    }

    public bool IsNull(Expression exp)
    {
      return false;
    }

    public ExpressionType TypeOf(Expression exp)
    {
      return ExpressionType.Int32;
    }

    public bool IsVariable(Expression exp)
    {
      return exp.IsVar;
    }

    public bool IsSlackVariable(Expression exp)
    {
      return exp.IsSlackVar;
    }

    public bool IsConstant(Expression exp)
    {
      return exp.IsConst;
    }

    public bool IsWritableBytes(Expression exp)
    {
      return false;
    }

    public bool IsUnaryExpression(Expression exp)
    {
      return exp.IsUnary;
    }

    public bool IsBinaryExpression(Expression exp)
    {
      return exp.IsBinary;
    }

    public bool IsSizeOf(Expression exp)
    {
      return false;
    }

    public object Variable(Expression exp)
    {
      return exp;
    }

    public object Constant(Expression exp)
    {
      return exp;
    }

    public bool TrySizeOf(Expression type, out int value)
    {
      value = -1;
      return false;
    }

    public System.Collections.Generic.ISet<Expression> VariablesIn(Expression exp)
    {
      return exp.Variables;
    }

    public bool TryValueOf<T>(Expression exp, ExpressionType aiType, out T value)
    {
      if (exp.IsConst)
      {
        int k;
        exp.Unpack(out k);
        IConvertible ic = k as IConvertible;
        if (ic != null)
        {
          try
          {
            value = (T)ic.ToType(typeof(T), null);
            return true;
          }
          catch
          { }
        }
        value = default(T);
        return false;
      }
      else
      {
        value = default(T);
        return false;
      }
    }

    public string NameOf(Expression exp)
    {
      return exp.ToString();
    }

    public Expression ToBooleanNormalizedForm(Expression exp)
    {
      return exp;
    }

    #endregion

    #region Privates
    static private ExpressionOperator ToExpressionOperator(UnaryOp un_op)
    {
      switch (un_op)
      {
        case UnaryOp.Minus:
          return ExpressionOperator.UnaryMinus;
        case UnaryOp.Not:
          return ExpressionOperator.Not;
        default:
          throw new Exception();
      }
    }

    static private ExpressionOperator ToExpressionOperator(BinaryOp bin_op)
    {
      switch (bin_op)
      {
        case BinaryOp.Add:
          return ExpressionOperator.Addition;

        case BinaryOp.And:
          return ExpressionOperator.And;

        case BinaryOp.Div:
          return ExpressionOperator.Division;

        case BinaryOp.Equal:
          return ExpressionOperator.Equal;
          
        case BinaryOp.Less:
          return ExpressionOperator.LessThan;

        case BinaryOp.LessEqual:
          return ExpressionOperator.LessEqualThan;

        case BinaryOp.GreaterEqual:
          return ExpressionOperator.GreaterEqualThan;

        case BinaryOp.NotEqual:
          return ExpressionOperator.NotEqual;

        case BinaryOp.Greater:
          return ExpressionOperator.GreaterThan;

        case BinaryOp.Mul:
          return ExpressionOperator.Multiplication;

        case BinaryOp.Or:
          return ExpressionOperator.Or;

        case BinaryOp.Sub:
          return ExpressionOperator.Subtraction;

        default:
          throw new Exception();
      }
    }
    #endregion
  }

  class Encoder : IExpressionEncoder<Expression>
  {
    #region IExpressionEncoder<Expression> Members

    public Expression FreshVariable<T>()
    {
      return new SlackVar();
    }

    public Expression ConstantFor(object value)
    {
      // In the example we handle just int
      if (value is Int32)
      {
        return new Const((Int32)value);
      }
      else
      {
        return null;
      }
    }

    public Expression CompoundExpressionFor(ExpressionType type, ExpressionOperator op, Expression left)
    {
      switch (op)
      {
        case ExpressionOperator.Not:
          return new UnaryExpression(UnaryOp.Not, left);

        case ExpressionOperator.UnaryMinus:
          return new UnaryExpression(UnaryOp.Minus, left);

        default:
          throw new InvalidOperationException();
      }
    }

    public Expression CompoundExpressionFor(ExpressionType type, ExpressionOperator op, Expression left, Expression right)
    {
      switch (op)
      {
        case ExpressionOperator.Addition:
          return new BinaryExpression(BinaryOp.Add, left, right);

        case ExpressionOperator.And:
          return new BinaryExpression(BinaryOp.And, left, right);

        case ExpressionOperator.Division:
          return new BinaryExpression(BinaryOp.Div, left, right);

        case ExpressionOperator.Equal:
          return new BinaryExpression(BinaryOp.Equal, left, right);

        case ExpressionOperator.LessThan:
          return new BinaryExpression(BinaryOp.Less, left, right);

        case ExpressionOperator.LessEqualThan:
          return new BinaryExpression(BinaryOp.LessEqual, left, right);

        case ExpressionOperator.GreaterEqualThan:
          return new BinaryExpression(BinaryOp.GreaterEqual, left, right);

        case ExpressionOperator.GreaterThan:
          return new BinaryExpression(BinaryOp.Greater, left, right);

        case ExpressionOperator.NotEqual:
          return new BinaryExpression(BinaryOp.NotEqual, left, right);

        case ExpressionOperator.Multiplication:
          return new BinaryExpression(BinaryOp.Mul, left, right);

        case ExpressionOperator.Or:
          return new BinaryExpression(BinaryOp.Or, left, right);

        case ExpressionOperator.Subtraction:
          return new BinaryExpression(BinaryOp.Sub, left, right);

        default:
          throw new Exception();
      }
    }

    public Expression Substitute(Expression source, Expression x, Expression exp)
    {
      throw new NotImplementedException();
    }

    #endregion
  }
}
