﻿using System;
using System.Collections.Generic;

using Microsoft.Research.AbstractDomains.Numerical;

namespace MiniClousot
{
  class MiniClousotMain
  {
    public static void Main(string[] args)
    {
      string filename;
      if (args.Length > 0 && TryParseCmdLine(args, out filename))
      {
        Parser parser = new Parser(new Scanner(filename));

        parser.Parse();
       
        if (parser.errors.count == 0)
        {
          Console.WriteLine("Parsed program\n{0}", parser.program.ToString());

          var decoder = new Decoder();
          var encoder = new Encoder();
          var top = new SubPoly<Expression>(decoder, encoder);
          var analysis = new Analyzer(top);

          DateTime start = DateTime.Now;

          Console.WriteLine("Analyzing...");
          analysis.Visit(parser.program);
          analysis.Check();

          Console.WriteLine("Time: {0}", DateTime.Now - start);

        }
      }
      else
      {
        Console.WriteLine("Use: miniclousot [-hints:CH | -hints:Oct | -hints:Full] [-reduction:Full | -reduction:Simplex | -reduction:Fast] filename");
      }
    }

    private static bool TryParseCmdLine(string[] args, out string filename)
    {
      filename = null;
      foreach (string arg in args)
      {
        switch (arg)
        {
          case "-hints:CH":
            SubPoly<Expression>.Inference = Hints.ConvexHull2D;
            break;

          case "-hints:Oct":
            SubPoly<Expression>.Inference = Hints.Octagons;
            break;

          case "-hints:Full":
            SubPoly<Expression>.Inference = Hints.CHOct;
            break;

          case "-hints:Standard":
            SubPoly<Expression>.Inference = Hints.Standard;
            break;

          case "-reduction:Full":
            SubPoly<Expression>.Algorithm = Reduction.Complete;
            break;

          case "-reduction:Simplex":
            SubPoly<Expression>.Algorithm = Reduction.Simplex;
            break;

          case "-reduction:Fast":
            SubPoly<Expression>.Algorithm = Reduction.Fast;
            break;

          default:
            filename = arg;
            break;
        }
      }

      return filename != null && !filename.StartsWith("-");
    }

  }

  class Analyzer : Visitor<SubPoly<Expression>>
  {
    Dictionary<Expression, SubPoly<Expression>> toCheck;     

    public Analyzer(SubPoly<Expression> initState)
    {
      this.toCheck = new Dictionary<Expression,SubPoly<Expression>>();
      this.result = initState;
    }

    public void Check()
    {
      foreach(var pair in this.toCheck)
      {
        var result = pair.Value.CheckIfHolds(pair.Key);
        Console.WriteLine("Assertion {0} is {1}", pair.Key, result.ToString());
      }
    }

    protected override void VisitAssume(Expression exp)
    {
      this.result = this.result.AssumeTrue(exp);
    }

    protected override void VisitAssert(Expression exp)
    {
      this.toCheck[exp] = this.Result.Duplicate();
      this.result = this.result.AssumeTrue(exp);
    }

    protected override void VisitAssign(Var x, Expression exp)
    {
      this.result.Assign(x, exp);
    }

    protected override void VisitNop()
    {
      // do nothing
    }

    protected override void VisitSeq(Statement s1, Statement s2)
    {
      this.Visit(s1);
      this.Visit(s2);
    }

    protected override void VisitIf(Expression exp, Statement s1, Statement s2)
    {
      SubPoly<Expression> prestate = this.Result.Duplicate();

      this.result = this.result.AssumeTrue(exp);
      this.Visit(s1);
      SubPoly<Expression> post1 = this.result;
      
      this.result = prestate.AssumeFalse(exp);
      this.Visit(s2);
      SubPoly<Expression> post2 = this.result;

      this.result = post1.Join(post2);
    }

    protected override void VisitWhile(Expression exp, Statement s1)
    {
      int widencount = 0;
      SubPoly<Expression> invariant = this.result.Duplicate();
      
      for (; ; )
      {
        this.result = this.result.AssumeTrue(exp);
        this.Visit(s1);
        SubPoly<Expression> widening = widencount++ >= 3 ? this.result.Widening(invariant) : this.result.Join(invariant);

        if (widening.LessEqual(invariant))
          break;

        invariant = widening;
        this.result = invariant.Duplicate();
      }

      this.result = this.result.AssumeFalse(exp);
    }
  }
}
